#ifdef AVR
#include <Arduino.h>
#else
#include <cstdint>
#include <cstring>
#endif

#include "ccsds_clcw.h"


// for debug:
/*#include <cstdio>
 #include <iostream>
 using namespace std;
 */

namespace CCSDS
  {
  
  uint32_t Clcw::create(const uint8_t u8_StatusField, const uint8_t u8_VirtualChannelID,
                        const bool b_NoRfAvail, const bool b_NoBitLock, const bool b_LockOut, const bool b_Wait, const bool b_Retransmit,
                        const uint8_t u8_FarmBCounter, const uint8_t u8_ReportValue)
  {
    uint32_t u32_CLCW = 0;
    
    u32_CLCW = ((uint32_t)(ClcwVersionNumber&0x3)<<29)
    | ((uint32_t)(u8_StatusField&0x7)<<26)
    | ((uint32_t)(COPinEffect&0x3)<<24)
    | ((uint32_t)(u8_VirtualChannelID&0x3f)<<18)
    | ((b_NoRfAvail?1:0)<<15) | ((b_NoBitLock?1:0)<<14) | ((b_LockOut?1:0)<<13) | ((b_Wait?1:0)<<12) | ((b_Retransmit?1:0)<<11)
    | ((uint32_t)(u8_FarmBCounter&0x3)<<9) | (uint32_t)(u8_ReportValue);
    
    return u32_CLCW;
  }
  
  
  
  int32_t Clcw::extract(uint8_t *pu8_StatusField, uint8_t *pu8_VirtualChannelID,
                        bool *pb_NoRfAvail, bool *pb_NoBitLock, bool *pb_LockOut, bool *pb_Wait, bool *pb_Retransmit,
                        uint8_t *pu8_FarmBCounter, uint8_t *pu8_ReportValue, const uint32_t u32_CLCW)
  {
    uint8_t u8_ClcwVersionNumber;
    
    if(pu8_StatusField)
      *pu8_StatusField=(uint8_t)((u32_CLCW>>26)&0x7);
    // COPinEffect not used
    if(pu8_VirtualChannelID)
      *pu8_VirtualChannelID=(uint8_t)((u32_CLCW>>18)&0x3f);
    
    if(pb_NoRfAvail)
      *pb_NoRfAvail=((u32_CLCW>>15)&0x1)?true:false;
    if(pb_NoBitLock)
      *pb_NoBitLock=((u32_CLCW>>14)&0x1)?true:false;
    if(pb_LockOut)
      *pb_LockOut=((u32_CLCW>>13)&0x1)?true:false;
    if(pb_Wait)
      *pb_Wait=((u32_CLCW>>12)&0x1)?true:false;
    if(pb_Retransmit)
      *pb_Retransmit=((u32_CLCW>>11)&0x1)?true:false;
    
    if(pu8_FarmBCounter)
      *pu8_FarmBCounter=(uint8_t)((u32_CLCW>>9)&0x3);
    if(pu8_ReportValue)
      *pu8_ReportValue=(uint8_t)(u32_CLCW&0xff);
    
    // version number check (last - to
    u8_ClcwVersionNumber=(uint8_t)((u32_CLCW>>29)&0x3);
    if(u8_ClcwVersionNumber!=ClcwVersionNumber)
      return -1;
    
    return 0;
  }
  





}
