#ifndef _CCSDS_SPACEPACKET_H_
#define _CCSDS_SPACEPACKET_H_

/****************************************************************/
/* SpacePackets according to                                    */
/*                                                              */
/*  - CCSDS 133.0-B-2, - Space Packet Protocol                  */
/*    https://public.ccsds.org/Pubs/133x0b2e1.pdf               */
/*                                                              */
/* Limitations:                                                 */ 
/*  - CCSDS secondary header format is not supportd             */
/*                                                              */
/****************************************************************/

#include "configCCSDS.h"

#ifdef configSP_MAX_DATA_SIZE
#define SP_MAX_DATA_SIZE configSP_MAX_DATA_SIZE
#else 
#define SP_MAX_DATA_SIZE 496
#endif






#define SP_HEADER_SIZE     6
#define SP_MAX_TOTAL_SIZE  (SP_HEADER_SIZE+SP_MAX_DATA_SIZE)



namespace CCSDS 
{


class SpacePacket
{
public:
  enum PacketType { TM = 0, TC = 1 };
  enum SequenceFlags { ContinuationSegment = 0x0, FirstSegment = 0x1, LastSegment = 0x2, Unsegmented = 0x3 };

  typedef void (TSpCallback)(void *p_SpContext, const uint8_t u8_PacketType,
                             const uint8_t u8_SequenceFlags, const uint16_t u16_APID,
                             const uint16_t u16_SequenceCount, const bool b_SecHeader,
                             const uint8_t *pu8_PacketData, const uint16_t u16_PacketDataLength);
  
private:
  const static uint8_t SpPacketVersion = 0;
  const static uint8_t PrimaryHdrSize = SP_HEADER_SIZE;
    
  uint32_t mu32_Index;
  uint8_t mu8_PacketVersionNumber;
  enum PacketType me_PacketType;
  bool mb_SecHdrFlag;
  uint16_t mu16_APID;
  enum SequenceFlags me_SequenceFlags;
  uint16_t mu16_PacketSequenceCount;
  uint16_t mu16_PacketDataLength;
  uint8_t au8_PacketData[SP_MAX_DATA_SIZE];
  
  bool mb_Overflow;
  uint16_t mu16_SyncErrorCount;
  uint16_t mu16_OverflowErrorCount;
  
  void *mp_SpContext;
  TSpCallback *mp_SpCallback;


public:
  SpacePacket(void *p_SpContext = NULL, TSpCallback *mp_SpCallback = NULL);
  
  void setCallback(void *p_SpContext, TSpCallback *mp_SpCallback);
  
  
  // SP generation
  static uint32_t create(uint8_t *pu8_Buffer, const uint32_t u32_BufferSize, 
                 const PacketType e_PacketType, const SequenceFlags e_SequenceFlags, const uint16_t u16_APID, const uint16_t u16_SequenceCount, 
                 const uint8_t *pu8_PacketData, const uint16_t u16_PacketDataLength);   // without secondary header

  static uint32_t create(uint8_t *pu8_Buffer, const uint32_t u32_BufferSize, 
                 const PacketType e_PacketType, const SequenceFlags e_SequenceFlags, const uint16_t u16_APID, const uint16_t u16_SequenceCount, 
                 const uint8_t *pu8_SecondaryHeaderData, const uint16_t u16_SecondaryHeaderLength,    
                 const uint8_t *pu8_PacketData, const uint16_t u16_PacketDataLength);

  static uint32_t createIdle(uint8_t *pu8_Buffer, const uint32_t u32_BufferSize, 
                 const uint16_t u16_SequenceCount, const uint16_t u16_TargetPacketSize);

  // sp processing
  int32_t process(const uint8_t *pu8_Buffer, const uint32_t u32_BufferSize);
  int32_t reset(void);
  
  uint16_t getSyncErrorCount(void);
  uint16_t getOverflowErrorCount(void);
  void clearErrorCounters(void);

private:
  static int32_t _create_primary_header(uint8_t *pu8_Buffer,  
                 const PacketType e_PacketType, const SequenceFlags e_SequenceFlags, const uint16_t u16_APID, const uint16_t u16_SequenceCount, const bool b_SecHeader, 
                 const uint32_t u32_PacketDataLength); 
};

}

#endif // _CCSDS_SPACEPACKET_H_
