#ifdef AVR
#include <Arduino.h>
#else
#include <cstdint>
#include <cstring>
#endif

#include "ccsds_transferframe.h"


// for debugging
//using namespace std;
//#include <iostream>

/*#include <iomanip>
void displaybuffer(const uint8_t *pu8_Buffer, const uint16_t u16_BufferSize)
{
    cout << "["; // << hex << setw(8) << setfill('0') << (uint32_t)pu8_Buffer << ", ";
  cout << dec << (uint32_t)u16_BufferSize << ": ";
  for(uint16_t i=0; i<u16_BufferSize; i++)
  {
    if(i%16==0 && i>0)
      cout << endl;
    cout << hex << setw(2) << setfill('0') << (uint32_t)pu8_Buffer[i] << ' ';
  }
  cout << ']' << endl;
}*/



namespace CCSDS 
{

  Transferframe::Transferframe(void)
  {
    mu16_Index = 0;
    mu16_FrameLength = 0;
    mb_Sync = true;
    mu16_SyncErrorCount = 0;
    mu16_ChecksumErrorCount = 0;
    mu16_OverflowErrorCount = 0;
  }



  void Transferframe::setSync(void)
  {
    mu16_Index=SyncSize;
    mb_Sync=true;
  }



  void Transferframe::process(const uint8_t *pu8_Data, const uint16_t u16_DataSize)
  {
    uint16_t u16_MaxTfSize = _getMaxTfSize();
    uint8_t *pu8_Buffer = _getTfBufferAddr();
    uint16_t u16_PrimaryHdrSize = _getPrimaryHeaderSize();
    bool b_Valid = true; 

 //   cout << "Tf::process(); max size: " << u16_MaxTfSize << endl;
    uint8_t au8_Sync[SyncSize]={0x1a, 0xcf, 0xfc, 0x1d};

    for(uint16_t i=0; i<u16_DataSize; i++)
    {
      if(mu16_Index<SyncSize)
      {
        if(pu8_Data[i] == au8_Sync[mu16_Index])
        {
          mu16_Index++;
        }
        else 
        {
          mb_Sync=false;
          if(pu8_Data[i]==au8_Sync[0])
            mu16_Index=1;
          else
            mu16_Index=0;
        }
  //      cout << mu16_Index;
      } 
      else if((mu16_Index<SyncSize+u16_PrimaryHdrSize) || (mu16_Index<SyncSize+mu16_FrameLength+1))
      {
        if(mu16_Index-SyncSize<u16_MaxTfSize)
          pu8_Buffer[mu16_Index-SyncSize]=pu8_Data[i];
        mu16_Index++;       
    //    cout << '.';
      }
 
      if(mu16_Index==SyncSize)
      {
        if(!mb_Sync && mu16_SyncErrorCount<0xffff)
          mu16_SyncErrorCount++;
        mb_Sync=true;   
     //   cout << '>';
      }
        
      if(mu16_Index==SyncSize+(uint32_t)u16_PrimaryHdrSize)
      {
      //  cout << 'H';
      //displaybuffer(pu8_Buffer, u16_PrimaryHdrSize);
        _getFrameLength();
        if(mu16_FrameLength+1>_getMaxTfSize())
        {
          mb_Sync=false;
          mu16_Index=0;
          mu16_FrameLength=0;
          if(mu16_OverflowErrorCount<0xffff)
            mu16_OverflowErrorCount++;
        }
      }
        
      if((mu16_Index>SyncSize+(uint32_t)u16_PrimaryHdrSize) && (mu16_Index>=SyncSize+mu16_FrameLength+1))
      {
      //  cout << "C: ";
      //  displaybuffer(pu8_Buffer, mu16_FrameLength+1);
      //  cout << endl << "  ";

#if TF_USE_FECF == 1
          b_Valid = _checkCRC();
          if(!b_Valid && (mu16_ChecksumErrorCount<0xffff))
            mu16_ChecksumErrorCount++;
#endif

        if(b_Valid)
          _processFrame();
        mu16_Index=0;
        mu16_FrameLength=0;
      }
    }
  }



  uint16_t Transferframe::getSyncErrorCount(void)
  {
    return mu16_SyncErrorCount + ((mu16_SyncErrorCount<0xffff && !mb_Sync)?1:0);
  }
 


  uint16_t Transferframe::getChecksumErrorCount(void)
  {
    return mu16_ChecksumErrorCount;
  }


  uint16_t Transferframe::getOverflowErrorCount(void)
  {
    return mu16_OverflowErrorCount;
  }



  void Transferframe::clearErrorCounters(void)
  {
    mu16_ChecksumErrorCount=0;
    mu16_SyncErrorCount=0;
    mu16_OverflowErrorCount=0;
    return;
  } 



  bool Transferframe::_checkCRC(void)
  {
    uint8_t *pu8_Buffer = _getTfBufferAddr();
 //   uint16_t u16_PrimaryHdrSize = _getPrimaryHeaderSize();
    uint16_t u16_FrameCRC;
    uint16_t u16_CalcCRC;
 
    u16_FrameCRC = (uint16_t)(pu8_Buffer[mu16_FrameLength+1-2]<<8) | (uint16_t)pu8_Buffer[mu16_FrameLength+1-1];
    u16_CalcCRC = Transferframe::calcCRC(pu8_Buffer, mu16_FrameLength+1-2);

 //   printf("[0x%04x/0x%04x]", u16_FrameCRC, u16_CalcCRC);

    return (u16_FrameCRC == u16_CalcCRC)?true:false; 
  }



  uint16_t Transferframe::calcCRC(const uint8_t *pu8_Buffer, const uint16_t u16_BufferSize)
  {
    uint16_t u16_CRC=0xffff;
    uint16_t u16_DataStreamXorBit15;

    // polynom: G(X) = X^16 + X^12 + X^5 + 1 

    for(uint16_t u16_BytePos = 0; u16_BytePos<u16_BufferSize; u16_BytePos++)
    {
      for(uint32_t u8_BitPos = 0; u8_BitPos<8; u8_BitPos++)
      {
        u16_DataStreamXorBit15 = ((pu8_Buffer[u16_BytePos]>>(7-u8_BitPos))&0x1) ^ ((u16_CRC>>15)&0x1);
        u16_CRC = (uint16_t)((u16_CRC<<1) ^ ((u16_DataStreamXorBit15<<12) | (u16_DataStreamXorBit15<<5) | (u16_DataStreamXorBit15)));
      }
    }

    return u16_CRC; 
  }









/*Transferframe::Transferframe()
{ long i;
  int tmp = 0;
  for(i=0; i<256; i++)
  {
    tmp=0;
    if((i&  1)!= 0) tmp = tmp ^ 0x1021;
    if((i&  2)!= 0) tmp = tmp ^ 0x2042;
    if((i&  4)!= 0) tmp = tmp ^ 0x4084;  
    if((i&  8)!= 0) tmp = tmp ^ 0x8108;
    if((i& 16)!= 0) tmp = tmp ^ 0x1231;
    if((i& 32)!= 0) tmp = tmp ^ 0x2462;
    if((i& 64)!= 0) tmp = tmp ^ 0x48c4;
    if((i&128)!= 0) tmp = tmp ^ 0x9188;
    lookUpTable[i]=tmp;
  }
}

uint16_t Transferframe::ComputeCRC(unsigned char *buf, long len, long initValue)
{  long currentValue = initValue;
   for(int i=0; i<len; i++)
   {
     currentValue = ( ((currentValue <<8) & 0xFF00) ^
          lookUpTable[(((currentValue>>8)^buf[i]) & 0x00FF)]);
   }
   return ((unsigned short)(currentValue & 0xFFFF));
}
*/
}
