#ifndef _CCSDS_TRANSFERFRAME_TC_H_
#define _CCSDS_TRANSFERFRAME_TC_H_

/****************************************************************/
/* TC Transferframes according to                               */
/*                                                              */
/*  - CCSDS 231.0-B-3, - TC Synchronization and Channel Coding  */
/*    https://public.ccsds.org/Pubs/231x0b3.pdf                 */
/*  - CCSDS 232.0-B-3, - TC Space Data Link Protocol            */
/*    https://public.ccsds.org/Pubs/232x0b3.pdf                 */
/*                                                              */
/* Limitations:                                                 */ 
/*  - The TC segment header is not supported                    */
/*                                                              */
/* Remarks:                                                     */
/*  - the sync code 0x1acffc1d is not generated with the        */
/*    by the create() method.                                   */
/*                                                              */
/****************************************************************/

#include "configCCSDS.h"

#ifdef configTC_TF_MAX_SIZE  
#define TC_TF_MAX_SIZE configTC_TF_MAX_SIZE
#else
#define TC_TF_MAX_SIZE 508
#endif

#ifdef configTC_TF_PRIM_HEADER_SIZE
#define TC_TF_PRIM_HEADER_SIZE configTC_TF_PRIM_HEADER_SIZE
#else
#define TC_TF_PRIM_HEADER_SIZE 5
#endif



#include "ccsds_transferframe.h"


namespace CCSDS 
{

  typedef void (TTcCallback)(void *p_Context, const bool b_BypassFlag, const bool b_CtrlCmdFlag,
                           const uint16_t u16_SpacecraftID, const uint8_t u8_VirtualChannelID,
                           const uint8_t u8_FrameSeqNumber,
                           const uint8_t *pu8_Data, const uint16_t u16_DataSize);

  class TransferframeTc : public Transferframe
  {
  private:
    static const int TcTfVersionNumber = 0;
    const static uint8_t PrimaryHdrSize = TC_TF_PRIM_HEADER_SIZE;
    const static uint16_t MaxTfSize = TC_TF_MAX_SIZE;
    uint8_t mau8_Buffer[MaxTfSize];   

    void *mp_TcContext;
    TTcCallback *mp_TcCallback;

  public:
    TransferframeTc(void *p_TcContext = NULL, TTcCallback *mp_TcCallback = NULL);
    
    void setCallback(void *p_TcContext, TTcCallback *mp_TcCallback);

    
    // TC generation
    static uint32_t create(uint8_t *pu8_Buffer, const uint32_t u32_BufferSize, 
                           const bool b_BypassFlag, const bool b_CtrlCmFlag, 
                           const uint16_t u16_SpacecraftID, const uint8_t u8_VirtualChannelID, 
                           const uint8_t u8_FrameSeqNumber, 
                           const uint8_t *pu8_Data, const uint16_t u16_DataSize);
    
  private:
    static int32_t _createPrimaryHeader(uint8_t *pu8_Buffer,  
                           const bool b_BypassFlag, const bool b_CtrlCmdFlag, 
                           const uint16_t u16_SpacecraftID, const uint8_t u8_VirtualChannelID, 
                           const uint16_t u16_FrameLength, const uint8_t u8_FrameSeqNumber);

    int32_t _processFrame(void);

  private:
    inline uint16_t _getMaxTfSize(void);  
    inline uint8_t *_getTfBufferAddr(void);  
    inline uint16_t _getPrimaryHeaderSize(void);  
    inline void _getFrameLength(void);  
  };


}


#endif // _CCSDS_TRANSFERFRAME_TC_H_


