#ifndef _CCSDS_TRANSFERFRAME_TM_H_
#define _CCSDS_TRANSFERFRAME_TM_H_


/****************************************************************/
/* TM Transferframes according to                               */
/*                                                              */
/*  - CCSDS 131.0-B-3, - TM Synchronization and Channel Coding  */
/*    https://public.ccsds.org/Pubs/131x0b3e1.pdf               */
/*  - CCSDS 132.0-B-2, - TM Space Data Link Protocol            */
/*    https://public.ccsds.org/Pubs/132x0b2.pdf                 */
/*                                                              */
/* Limitations:                                                 */ 
/*  - The TM secondary header is not supported                  */
/*  - Randomization is not supported                            */
/*                                                              */
/****************************************************************/


#include "configCCSDS.h"

#ifdef configTM_TF_TOTAL_SIZE  
#define TM_TF_TOTAL_SIZE configTM_TF_TOTAL_SIZE
#else 
#define TM_TF_TOTAL_SIZE 508
#endif

#ifdef configTF_USE_OCF
#define TF_USE_OCF configTF_USE_OCF
#else
#define TF_USE_OCF 1
#endif


#include "ccsds_transferframe.h"


namespace CCSDS 
{

  typedef void (TTmCallback)(void *p_Context, const uint16_t u16_SpacecraftID, const uint8_t u8_VirtualChannelID,
                             const uint8_t u8_MasterChannelFrameCount, const uint8_t u8_VirtualChannelFrameCount,
                             const bool b_TFSecHdrFlag, const uint16_t u16_FirstHdrPtr,
                             const uint8_t *pu8_Data, const uint16_t u16_DataSize,
                             const uint32_t u32_OCF);


  class TransferframeTm : public Transferframe
  {
  private:
    static const int TmTfVersionNumber = 0;
    const static uint8_t PrimaryHdrSize = 6;
    const static uint8_t OcfSize = 4;
    const static uint16_t TfSize = TM_TF_TOTAL_SIZE;
    uint8_t mau8_Buffer[TfSize];   

    const static bool UseOCF = (TF_USE_OCF)?true:false;  // Operational Control Field (CLCW)

    void *mp_TmContext;
    TTmCallback *mp_TmCallback;

  public:    
    TransferframeTm(void *p_TmContext = NULL, TTmCallback *mp_TmCallback = NULL);
    
    void setCallback(void *p_TmContext, TTmCallback *mp_TmCallback);


  // TM generation
    static uint32_t create(uint8_t *pu8_Buffer, const uint32_t u32_BufferSize, 
                 const uint16_t u16_SpacecraftID, const uint8_t u8_VirtualChannelID,
                 const uint8_t u8_MasterChannelFrameCount, const uint8_t u8_VirtualChannelFrameCount, 
                 const uint16_t u16_FirstHdrPtr,
                 const uint8_t *pu8_Data, const uint16_t u16_DataSize,
                 const uint32_t u32_OCF = 0);

    static uint32_t createIdle(uint8_t *pu8_Buffer, const uint32_t u32_BufferSize,
                 const uint16_t u16_SpacecraftID, const uint8_t u8_VirtualChannelID,
                 const uint8_t u8_MasterChannelFrameCount, const uint8_t u8_VirtualChannelFrameCount,
                 const uint32_t u32_OCF = 0);

  private:
    static int32_t _createPrimaryHeader(uint8_t *pu8_Buffer,  
                 const uint16_t u16_SpacecraftID, const uint8_t u8_VirtualChannelID, const bool b_OcfFlag,
                 const uint8_t u8_MasterChannelFrameCount, const uint8_t u8_VirtualChannelFrameCount, 
                 const bool b_TFSecHdrFlag, const bool b_SyncFlag, const bool b_PacketOrderFlag, 
                 const uint8_t u8_SegLengthID, const uint16_t u16_FirstHdrPtr); 

    int32_t _processFrame(void);

  private:
    inline uint16_t _getMaxTfSize(void);  
    inline uint8_t *_getTfBufferAddr(void);  
    inline uint16_t _getPrimaryHeaderSize(void);  
    inline void _getFrameLength(void);  
  };


}


#endif // _CCSDS_TRANSFERFRAME_TM_H_


